"""
To clarify:

Old CS mapping:
geo_coordinates.nc:latitude -> coord:latitude

New ACRI mapping
geo_coordinates.nc:latitude -> /quality/latitude

DataTree
S03OLCEFR.quality.oa01_radiance_unc.coordinates
'latitude longitude'


"""

import json
from pathlib import Path

import numpy
import numpy as np
from eopf.store.mapping_manager import EOPFMappingManager
from IPython.display import JSON
from xarray import DataTree

import sentineltoolbox.api as stb
from sentineltoolbox.datatree_utils import get_array
from sentineltoolbox.readers.resources import get_resource_path, load_resource_file
from sentineltoolbox.resources.reference import PRODUCT
from sentineltoolbox.tools.stb_dump_product import (
    convert_datatree_to_structure_str,
    convert_mapping_to_datatree,
)
from sentineltoolbox.writers.json import DataTreeJSONEncoder

COORDS_NAMESPACES = {
    "S03OLCEFR": {
        "geo_coordinates.nc:columns": "coords:columns",
        "geo_coordinates.nc:latitude": "coords:latitude",
        "geo_coordinates.nc:longitude": "coords:longitude",
        "geo_coordinates.nc:rows": "coords:rows",
        "instrument_data.nc:bands": "coords:bands2",
        "instrument_data.nc:detector_index": "coords:detector_index",
        "instrument_data.nc:detectors": "coords:detectors",
        "instrument_data.nc:frame_offset": "coords:frame_offset",
        "removed_pixels.nc:detector_index": "coords:orphans:detector_index",
        "removed_pixels.nc:frame_offset": "coords:orphans:frame_offset",
        "removed_pixels.nc:latitude": "coords:orphans:latitude",
        "removed_pixels.nc:longitude": "coords:orphans:longitude",
        "removed_pixels.nc:removed_pixels": "coords:orphans:removed_pixels",
        "tie_geo_coordinates.nc:latitude": "coords:tie_point:latitude",
        "tie_geo_coordinates.nc:longitude": "coords:tie_point:longitude",
        "tie_geo_coordinates.nc:tie_columns": "coords:tp_columns",
        "tie_geo_coordinates.nc:tie_rows": "coords:tp_rows",
        "tie_meteo.nc:tie_pressure_levels": "coords:tie_point:pressure_level",
        "tie_meteo.nc:wind_vectors": "coords:tie_point:wind_vector",
        "time_coordinates.nc:time_stamp": "coords:time_stamp",
    },
}


COORDS_DEF: dict[tuple[str, tuple[int, ...], str], str] = {
    # /conditions/orphans/detector_index, /conditions/orphans/frame_offset, /conditions/orphans/sza, /measurements/orphans/oa<NUM>_radiance, /quality/orphans/oa<NUM>_radiance_unc, /quality/orphans/quality_flags # noqa: E501
    ("altitude", (4092, 150), "float32"): "coords:orphans:altitude",
    # /conditions/image/detector_index, /conditions/image/frame_offset
    ("altitude", (4092, 4865), "float32"): "coords:altitude",
    # /conditions/geometry/oaa, /conditions/geometry/oza, /conditions/geometry/saa, /conditions/geometry/sza
    ("latitude", (4092, 77), "float64"): "coords:tie_point:latitude",
    # /conditions/orphans/detector_index, /conditions/orphans/frame_offset, /conditions/orphans/sza, /measurements/orphans/oa<NUM>_radiance, /quality/orphans/oa<NUM>_radiance_unc, /quality/orphans/quality_flags # noqa: E501
    ("latitude", (4092, 150), "float64"): "coords:orphans:latitude",
    # /measurements/oa<NUM>_radiance, /quality/oa<NUM>_radiance_unc, /quality/quality_flags, /conditions/image/detector_index, /conditions/image/frame_offset # noqa: E501
    ("latitude", (4092, 4865), "float64"): "coords:latitude",
    # /conditions/geometry/oaa, /conditions/geometry/oza, /conditions/geometry/saa, /conditions/geometry/sza
    ("longitude", (4092, 77), "float64"): "coords:tie_point:longitude",
    # /conditions/orphans/detector_index, /conditions/orphans/frame_offset, /conditions/orphans/sza, /measurements/orphans/oa<NUM>_radiance, /quality/orphans/oa<NUM>_radiance_unc, /quality/orphans/quality_flags # noqa: E501
    ("longitude", (4092, 150), "float64"): "coords:orphans:longitude",
    # /measurements/oa<NUM>_radiance, /quality/oa<NUM>_radiance_unc, /quality/quality_flags, /conditions/image/detector_index, /conditions/image/frame_offset # noqa: E501
    ("longitude", (4092, 4865), "float64"): "coords:longitude",
    # /conditions/meteorology/atmospheric_temperature_profile
    ("pressure_level", (25,), "float32"): "coords:pressure_level",
    # /measurements/oa<NUM>_radiance
    ("time_stamp", (4092,), "datetime64[ns]"): "coords:time_stamp",
}


def f() -> None:
    mappings = EOPFMappingManager()
    mapping, _ = mappings.parse_mapping(product_type="S03OLCEFR", processing_version="TODO")
    JSON(mapping)

    prod = convert_mapping_to_datatree(mapping)
    # PRODUCT.map()

    print(convert_datatree_to_structure_str(prod, details=True))

    S03OLCEFR = PRODUCT["S03OLCEFR"]
    S03OLCEFR.measurements.oa01_radiance


simplified = {}
simplified["S03OLCEFR"] = load_resource_file(
    "simplified_mappings/S3_OL_1_mapping.json",
    module="sentineltoolbox.conversion.conf",
)


def _extract_coord() -> None:
    """
    if "/" in dpr_tgt:
        ns = ":".join(dpr_tgt.split("/"))
        print(f"{dpr_tgt!r}:'coords{ns}'")

    """


def f2(ptype: str, cache_dir: str | Path = "~/DATA/Products") -> None:
    """
     Mapping:
     {
         "short_name": "oa01_radiance",
         "transform": {
             "dimensions": ["rows", "columns"],
             "attributes": {
                 "coordinates": "latitude longitude",
                 "dimensions": "rows columns",
                 "dtype": "<u2",
             }
             "mask_and_scale": {
                    "valid_min": -180000000,
                    "fill_value": -2147483648,
                    "valid_max": 180000000,
                    "scale_factor": 1e-06
                },
         },
     }


    xdataarray.attrs:

    {
        "_ARRAY_DIMENSIONS": ["rows", "columns"],
        "coordinates": "time_stamp latitude longitude",
        "short_name": "oa01_radiance",
        "standard_name": "toa_upwelling_spectral_radiance",
        "units": "mW.m-2.sr-1.nm-1",
        "valid_max": 893.550000693649,
        "valid_min": 0.0,
        "_io_config": {
            "scale_factor": 0.013634907081723213,
            "add_offset": 0.0,
            "valid_min": 0,
            "valid_max": 65534,
            "fill_value": 65535,
            "dtype": dtype("uint16"),
        },
        "description": "TOA radiance for OLCI acquisition band oa01",
    }


    xdataarray.encoding

    {"_FillValue": 65535, "scale_factor": 0.013634907081723213, "add_offset": 0.0, "dtype": dtype("uint16")}
    """
    scale_factor_mode = "ignore"

    cache_dir = Path(cache_dir).expanduser()
    if ptype.startswith("S03"):
        ref_prod_path = f"s3://s3-input/Products/{ptype}*.zarr"
    elif ptype.startswith("S02"):
        ref_prod_path = f"s3://s2-input/Products/{ptype}*.zarr"
    else:
        ref_prod_path = "XXXXX"

    try:
        full_prod = stb.open_datatree(ref_prod_path, cache=True, local_copy_dir=cache_dir)
        # full_prod_raw = stb.open_datatree(ref_prod_path, cache=True, local_copy_dir=cache_dir, decode_cf=False)
    except FileNotFoundError:
        full_prod = None
        # full_prod_raw = None

    coord_ns = COORDS_NAMESPACES.get(ptype, {})
    mappings = EOPFMappingManager()
    mapping_expanded, _ = mappings.parse_mapping(product_type=ptype, processing_version="TODO")
    simp = simplified.get(ptype, {})
    # chunk_sizes = simp.get("chunk_sizes")
    ref_product: DataTree = PRODUCT[ptype]
    data_mapping = simp.get("data_mapping", {})
    map_legacy_to_dpr = {}
    for gr_path, legacy_files in data_mapping.items():
        for legacy_file, variables in legacy_files.items():
            for legacy_name, dpr_name in variables:
                src_path = f"{legacy_file}:{legacy_name}"
                target_path = f"{gr_path}/{dpr_name}"
                map_legacy_to_dpr[src_path] = target_path
    real_mapping = load_resource_file(f"{ptype}.json", module="eopf.store.mapping")

    RTOL = 1e-8
    ATOL = 1e-8
    for data in real_mapping.get("data_mapping"):
        # 'target_path': 'coords:rows', 'source_path': 'geo_coordinates.nc:rows'

        src = data.get("source_path")
        map_tgt = data.get("target_path")
        dpr_tgt = map_legacy_to_dpr.get(src, map_tgt)

        if map_tgt.startswith("coords:"):
            data["target_path"] = coord_ns.get(src, map_tgt)
        else:
            data["target_path"] = dpr_tgt

        if isinstance(full_prod, DataTree):
            try:
                var_metadata = ref_product[dpr_tgt]
                var_prod = get_array(full_prod, dpr_tgt)
            except KeyError:
                pass
            else:
                io = var_metadata.attrs.get("_io_config", var_metadata.attrs)
                use_ns = ""
                coordinates: list[str] = []
                for coord_name, coord in var_prod.coords.items():
                    coord_name = str(coord_name)
                    coord_id: tuple[str, tuple[int, ...], str] = (coord_name, coord.shape, coord.dtype.name)
                    # coord_path = "/".join(dpr_tgt.split("/")[:-1] + [coord_name])
                    coordinates.append(COORDS_DEF.get(coord_id, coord_name))
                    if coord_id in COORDS_DEF:
                        ns = COORDS_DEF[coord_id]
                        parts = ns.split(":")
                        if len(parts) > 2 and not data["target_path"].startswith("coords:"):
                            use_ns = ns.split(":")[1]
                if coordinates and not data["target_path"].startswith("coords:"):
                    coordinates_str = " ".join([c.split(":")[-1] for c in coordinates])
                    data.setdefault("transform", {}).setdefault("attributes", {})["coordinates"] = coordinates_str
                if use_ns:
                    data["coords_namespace"] = use_ns
                else:
                    try:
                        del data["coords_namespace"]
                    except KeyError:
                        pass

                for field in ("scale_factor", "add_offset", "valid_min", "valid_max", "fill_value", "dtype"):
                    if field in io:
                        value = io[field]
                        if scale_factor_mode == "remove":
                            # see https://gitlab.eopf.copernicus.eu/cpm/eopf-cpm/-/issues/547
                            # REMOVE ALL add_offset and scale_factor of all variables
                            if field == "add_offset":  # and numpy.isclose(value, 0):
                                try:
                                    del data["transform"]["mask_and_scale"][field]
                                except KeyError:
                                    pass
                                continue
                            elif field == "scale_factor":  # and numpy.isclose(value, 1):
                                try:
                                    del data["transform"]["mask_and_scale"][field]
                                except KeyError:
                                    pass
                                continue
                        elif scale_factor_mode == "update":
                            # REMOVE add_offset==0 and scale_factor==1
                            if field == "add_offset" and numpy.isclose(value, 0, rtol=RTOL, atol=ATOL):
                                try:
                                    del data["transform"]["mask_and_scale"][field]
                                except KeyError:
                                    pass
                                continue
                            elif field == "scale_factor" and numpy.isclose(value, 1):
                                try:
                                    del data["transform"]["mask_and_scale"][field]
                                except KeyError:
                                    pass
                                continue
                        else:
                            if field == "add_offset" or field == "scale_factor":
                                continue

                        if field == "fill_value" and value is numpy.nan:
                            continue

                        # attributes
                        if field == "dtype":
                            # storage type
                            old_dtype = np.dtype(data.get("transform", {}).get("attributes", {}).get(field))
                            if np.dtype(value) != old_dtype:
                                data.setdefault("transform", {}).setdefault("attributes", {})[field] = np.dtype(
                                    value,
                                ).str
                            # array type: prod_var.dtype.str
                        else:
                            data.setdefault("transform", {}).setdefault("mask_and_scale", {})[field] = value
                # print(f"fix {map_tgt}")

    mapping_path_dir = get_resource_path("mapping", module="eopf.store")
    mapping_path = mapping_path_dir / f"{ptype}_new.json"
    with open(mapping_path, "w") as json_fp:
        json.dump(real_mapping, json_fp, indent=4, cls=DataTreeJSONEncoder)


f2("S03OLCEFR")
